/**
 ******************************************************************************
 * @file           : main.c
 * @author         : Auto-generated by STM32CubeIDE
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * Copyright (c) 2022 STMicroelectronics.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

#include <stdint.h>
#include <stdio.h>

#if !defined(__SOFT_FP__) && defined(__ARM_FP)
  #warning "FPU is not initialized, but the project is compiling for an FPU. Please initialize the FPU before use."
#endif

//Debug Exception and Monitor Control Register base address
#define DEMCR                 *((volatile uint32_t*) 0xE000EDFCU )

/* ITM register addresses */
#define ITM_STIMULUS_PORT0    *((volatile uint32_t*) 0xE0000000 )
#define ITM_TRACE_EN          *((volatile uint32_t*) 0xE0000E00 )

/* Send a char through ITM */
void ITM_SendChar(uint8_t ch)
{
  // read FIFO status in bit [0]:
  while(!(ITM_STIMULUS_PORT0 & 1));

  //Write to ITM stimulus port0
  ITM_STIMULUS_PORT0 = ch;
}

/* Override low-level _write system call */
int _write(int file, char *ptr, int len)
{
  int DataIdx;

  for (DataIdx = 0; DataIdx < len; DataIdx++)
  {
    ITM_SendChar(*ptr++);
  }
  return len;
}

/* Helper function */
void DumpExceptionRegister(uint32_t* pMSP)
{
  printf(" MSP = %p\n", pMSP);
  printf("  R0 = 0x%lx\n", pMSP[0]);  // May have argument of function
  printf("  R1 = 0x%lx\n", pMSP[1]);  // May have argument of function
  printf("  R2 = 0x%lx\n", pMSP[2]);  // May have argument of function
  printf("  R3 = 0x%lx\n", pMSP[3]);  // May have argument of function
  printf(" R12 = 0x%lx\n", pMSP[4]);  // IP holds an intermediate value of a calculation
  printf("  LR = 0x%lx\n", pMSP[5]);  // Address of the next instruction before the exception
  printf("  PC = 0x%lx\n", pMSP[6]);  // CPU was executing the instruction at PC
  printf("xPSR = 0x%lx\n", pMSP[7]);  // Status of system before execution at PC completes
}

/* Implement Fault Handlers */
void HardFault_Handler_main(uint32_t* pMSP) {
  printf("Exception: Hard Fault\n");
  DumpExceptionRegister(pMSP);

  while(1);
}

__attribute__ ((naked)) void HardFault_Handler(void) {
  __asm volatile("MRS R0, MSP"); // get the Stack Pointer right after Stacking procedure
  __asm volatile("B HardFault_Handler_main"); // call actual handler, passing R0 as the argument
}

void MemManage_Handler_main(uint32_t* pMSP) {
  printf("Exception: Memory Fault\n");
  DumpExceptionRegister(pMSP);

  while(1);
}

__attribute__ ((naked)) void MemManage_Handler(void) {
  __asm volatile("MRS R0, MSP");
  __asm volatile("B MemManage_Handler_main");
}

void BusFault_Handler_main(uint32_t* pMSP) {
  printf("Exception: Bus Fault\n");
  DumpExceptionRegister(pMSP);

  while(1);
}

__attribute__ ((naked)) void BusFault_Handler(void) {
  __asm volatile("MRS R0, MSP");
  __asm volatile("B BusFault_Handler_main");
}

void UsageFault_Handler_main(uint32_t* pMSP) {
  printf("Exception: Usage Fault\n");
  DumpExceptionRegister(pMSP);

  uint32_t* pUFSR = (uint32_t*)0xE000ED2A;
  printf("UFSR = 0x%lx\n", *pUFSR & 0xFFFF);

  while(1);
}

__attribute__ ((naked)) void UsageFault_Handler(void) {
  __asm volatile("MRS R0, MSP");
  __asm volatile("B UsageFault_Handler_main");
}

int div(int x, int y) {
  return x/y;
}

int main(void)
{
  // define SHCSR register
  uint32_t* pSHCSR = (uint32_t *)0xE000ED24;

  // enable handlers
  *pSHCSR |= (1 << 16); // Memory Fault
  *pSHCSR |= (1 << 17); // Bus Fault
  *pSHCSR |= (1 << 18); // Usage Fault

  // enable Divide by Zero Trap
  uint32_t* pCCR = (uint32_t*)0xE000ED14;
  *pCCR |= (1 << 4); // Div by Zero

#if 1

  /* Fill a meaningless value */
  *(uint32_t*)0x20010000 = 0xFFFFFFFF;

  /* Set PC without the LSB bit being 1 means no Thumb State => Usage Fault: Invalid State */
//  void (*pFunc)(void) = (void*)0x20010000;

  /* Set PC with LSB being 1 to indicate Thumb State */
  void (*pFunc)(void) = (void*)0x20010001;

  /* The data in the address that PC points is not a valid instruction => Usage Fault: Undefined Instruction */
  pFunc();

#endif

#if 1
  div(10,0);
#endif

  /* Loop forever */
  for(;;);
}
